package com.raffaeleconforti.foreignkeydiscovery.conceptualmodels;

import org.deckfour.xes.model.XAttribute;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/**
 * Defines an instance of an entity with specific values for the attributes Can
 * store more than one value per attribute in the same instance but these are
 * only used for timestamp attributes. For other attributes only the first value
 * is used. Currently no information is kept on which value of non-timestamp
 * attribute corresponds to which value of a timestamp attribute.
 *
 * @author Viara Popova
 */
public class EntityInstance {
    private HashMap<String, List<XAttribute>> attribute_values; //the attribute name and the value(s)

    public EntityInstance() {
        this.attribute_values = new HashMap<String, List<XAttribute>>();
    }

    public void addAttribute(Attribute attr, XAttribute value) {
        List<XAttribute> list;
        if ((list = attribute_values.get(attr.getName())) == null) { //element exists
            list = new ArrayList<XAttribute>();
            attribute_values.put(attr.getName(), list);
        }
        list.add(value);
    }

    public void addAttribute(String attrName, XAttribute value) {
        List<XAttribute> list;
        if ((list = attribute_values.get(attrName)) == null) { //element exists
            list = new ArrayList<XAttribute>();
            attribute_values.put(attrName, list);
        }
        list.add(value);
    }

    public String findCreation(List<Attribute> attrs) { //attrs should be the set of timestamp attributes
        String date = null;
        for (Attribute attr : attrs) {
            List<XAttribute> newdateArr;
            if ((newdateArr = attribute_values.get(attr.getName())) != null) {
                for (XAttribute aNewdateArr : newdateArr) {
                    String newdate = aNewdateArr.toString();
                    if (date == null || date.compareTo(newdate) > 0)
                        date = newdate;
                }
            }
        }
        return date;
    }

    //find the values for a set of attributes
    //assumes that no two attributes of an entity have the same names
    public HashMap<Attribute, XAttribute> getValues(List<Attribute> attrs) {
        HashMap<Attribute, XAttribute> values = new HashMap<Attribute, XAttribute>();
        for (Attribute attr : attrs) {
            List<XAttribute> vals = attribute_values.get(attr.getName());
            if (vals == null)
                return null;
            values.put(attr, vals.get(0)); //only the first value if more than one
        }
        return values;
    }

    public HashMap<String, XAttribute> getValuesStr(List<Attribute> attrs) {
        HashMap<String, XAttribute> values = new HashMap<String, XAttribute>();
        for (Attribute attr : attrs) {
            List<XAttribute> val = attribute_values.get(attr.getName()); //only the first value if more than one
            if (val == null)
                return null;
            values.put(attr.getName(), val.get(0));
        }
        return values;
    }

    //find the values for foreign keys attributes of a given entity
    public HashMap<Attribute, XAttribute> getValues(List<ForeignKey> attrs, Entity e) {
        HashMap<Attribute, XAttribute> values = new HashMap<Attribute, XAttribute>();
        for (ForeignKey attr : attrs) { //over all attributes
            if (attr.getEntity().equals(e)) {
                List<XAttribute> vals = attribute_values.get(attr.getFKey().getName());
                if (vals == null)
                    return null;
                values.put(attr.getKey(), vals.get(0)); //only the first value if more than one
            }
        }
        return values;
    }

    public HashMap<String, XAttribute> getValuesKeysStr(List<ForeignKey> attrs) {
        HashMap<String, XAttribute> values = new HashMap<String, XAttribute>();
        for (ForeignKey attr : attrs) { //over all attributes
            List<XAttribute> vals = attribute_values.get(attr.getFKey().getName());
            if (vals == null)
                return null;
            values.put(attr.getKey().getName(), vals.get(0)); //only the first value if more than one
        }
        return values;
    }

    public XAttribute getValue(String attrName) {//Attribute attr){
        if (attribute_values.get(attrName) == null)
            return null;
        return attribute_values.get(attrName).get(0); //only the first value if more than one
    }

    public ArrayList<XAttribute> getValueArray(String attrName) {//Attribute attr){
        if (attribute_values.get(attrName) == null)
            return null;
        ArrayList<XAttribute> valArr = new ArrayList<XAttribute>();
        for (XAttribute val : attribute_values.get(attrName)) {
            valArr.add(val);
        }
        return valArr; //only the first value if more than one
    }

    public int getSize(Attribute attr) {
        List<XAttribute> list;
        if ((list = attribute_values.get(attr.getName())) == null) return 0;
        return list.size();
    }
}
