package com.raffaeleconforti.foreignkeydiscovery.conceptualmodels;

import org.deckfour.xes.model.XAttribute;

import java.util.*;

public class EntityInstances {//the instances of one entity

    private HashMap<HashMap<String, XAttribute>, Set<EntityInstance>> cache = new HashMap<HashMap<String, XAttribute>, Set<EntityInstance>>();

    private ArrayList<EntityInstance> instances;

    public EntityInstances() {
        this.instances = new ArrayList<EntityInstance>();
    }

    public ArrayList<EntityInstance> getInstances() {
        return this.instances;
    }

    public void addInstance(EntityInstance inst) {
        this.instances.add(inst);
    }

    public ValueList getValues(Attribute attr) {
        ValueList values = new ValueList();
        for (EntityInstance inst : instances) {
            XAttribute val = inst.getValue(attr.getName());
            if (val != null) {
                String s = val.toString();
                values.addValue(s);
            }
        }
        return values;
    }

    public ValueList getValues(List<Attribute> attrs) {
        ValueList values = new ValueList();
        if (attrs.size() == 0)
            return values;
        for (EntityInstance inst : instances) { //for each instance
            StringBuilder val = new StringBuilder();
            Integer attr = 0;
            while (attr < attrs.size()) {
                Attribute a = attrs.get(attr);
                if(inst.getValue(a.getName()) != null) {
                    String s = inst.getValue(a.getName()).toString();
                    val.append("|"); //delimiter between values, assumes that this symbol does not appear in the values
                    val.append(s);
                    attr++;
                }else {
                    attr++;
                }
            }
            if (val.length() > 0)
                values.addValue(val.toString());
        }
        return values;
    }

    //returns the instances that contain these values for the specified attributes
    //if attrValues contain the (composite) key attributes then the resulting set will contain only one instance
    public Set<EntityInstance> getInstances(HashMap<String, XAttribute> attrValues) {
        Set<EntityInstance> foundInstances;
//        if((foundInstances = cache.get(attrValues)) == null) {
        foundInstances = new HashSet<EntityInstance>();
        for (EntityInstance inst : instances) {
            Boolean match = true;
            for (Map.Entry<String, XAttribute> entry : attrValues.entrySet()) {
                if (inst.getValue(entry.getKey()) == null || !inst.getValue(entry.getKey()).equals(entry.getValue())) {
                    match = false;
                    break;
                }
            }
            if (match) {
                foundInstances.add(inst);
            }
        }
//            cache.put(attrValues, foundInstances);
//        }
        return foundInstances;
    }

}
