package com.raffaeleconforti.foreignkeydiscovery.histogram.impl;

import com.raffaeleconforti.foreignkeydiscovery.Cell;
import com.raffaeleconforti.foreignkeydiscovery.databasestructure.Column;
import com.raffaeleconforti.foreignkeydiscovery.databasestructure.Tuple;
import com.raffaeleconforti.foreignkeydiscovery.histogram.Histogram;



import java.util.*;

/**
 * Created by Raffaele Conforti on 17/10/14.
 */
public class DistributionHistogram implements Histogram {

    private HashMap<Cell, Set<Tuple<String>>> quantileDistributionHistogram = new HashMap<Cell, Set<Tuple<String>>>();
    private HashMap<Cell, Double> histogram = null;
    private int numberTuples = 0;
    private int numbersOfQuantiles;
    private Integer hashCode;

    public DistributionHistogram(Histogram histogram, TreeSet<Column> columns) {

        this.numbersOfQuantiles = histogram.getQuantiles();
//        Column column = columns.iterator().next();

//        HashMap<String, Integer> columnNames = new HashMap<String, Integer>();
        Integer pos = 0;
        for(Column c : columns) {
//            columnNames.put(c.getColumnName(), pos);
            pos++;
        }

        Tuple<String>[] setTuples = createTuples(columns);
        this.numberTuples = setTuples.length;
        Arrays.sort(setTuples);

        for (Map.Entry<Cell, Set<Tuple<String>>> entry : histogram.getQuantileHistogram().entrySet()) {

            Cell cell = entry.getKey();
            Set<Tuple<String>> value = entry.getValue();
            Set<Tuple<String>> set = new HashSet<Tuple<String>>();

            for (Tuple<String> tuple : value) {
                if (Arrays.binarySearch(setTuples, tuple) > -1) {
                    set.add(tuple);
                }
            }

            quantileDistributionHistogram.put(cell, set);

        }

    }

    private Tuple<String>[] createTuples(Set<Column> columns) {
        Column[] arrayColumns = columns.toArray(new Column[columns.size()]);

        Tuple<String>[] tuples = new Tuple[arrayColumns[0].getColumnValues().getValues().length];

        for(int i = 0; i < tuples.length; i++) {
            String[] elements = new String[arrayColumns.length];
            for (int j = 0; j < arrayColumns.length; j++) {
                elements[j] = arrayColumns[j].getColumnValues().getValues()[i];
            }
            tuples[i] = new Tuple<String>(elements);
        }

        return  tuples;

    }

    @Override
    public HashMap<Cell, Double> getHistogram() {
        if(histogram == null) {
            histogram = new HashMap<Cell, Double>();
            for(Map.Entry<Cell, Set<Tuple<String>>> entry : quantileDistributionHistogram.entrySet()) {
                histogram.put(entry.getKey(), ((double) entry.getValue().size() / (double) numberTuples));
            }
        }
        return histogram;
    }

    @Override
    public HashMap<Cell, Set<Tuple<String>>> getQuantileHistogram() {
        return quantileDistributionHistogram ;
    }

    @Override
    public int getQuantiles() {
        return numbersOfQuantiles;
    }

    @Override
    public int hashCode() {
        if(hashCode == null) {
            hashCode = quantileDistributionHistogram.hashCode();
        }
        return hashCode;
    }

    @Override
    public boolean equals(Object o) {
        if(o instanceof DistributionHistogram) {
            DistributionHistogram dh = (DistributionHistogram) o;
            return dh.quantileDistributionHistogram.equals(quantileDistributionHistogram);
        }
        return false;
    }

    @Override
    public String toString() {
        return quantileDistributionHistogram.toString();
    }
}
